# Native imports
import ctypes
import ctypes.wintypes

# Third-party imports
import asyncio

# Custom imports
from wizsdk.constants import keycode_map

user32 = ctypes.windll.user32


class Keyboard:
    """
    Keyboard class.
    Sends events directly to the client. Must pass in ``window_handle`` of the client to use.
    """

    def __init__(self, window_handle):
        # super().__init__(window_handle)
        self.window_handle = window_handle
        self.key_tasks = {}

    async def hold_key(self, key, seconds=0.1):
        """
        Hold down a key for an amount of time. The key is sent directly to the client, the client does not need to be in focus.
        
        Args:
            key (str): The key to hold down. 
                "TAB", "ENTER", "ALT", "ESC", "SPACE" and others are also accepted as special keys
            seconds (int, optional): duration to hold for
        """
        self.key_down(key)
        await asyncio.sleep(seconds)
        self.key_up(key)

    async def send_key(self, key, seconds=0.1):
        """
        Alias for ``hold_key``
        """
        await self.hold_key(key, seconds)

    def key_down(self, key):
        """
        Hold down a key. Call ``key_up`` to release.
        
        Args:
            key (str): The key to hold down
        """
        self.key_tasks[key] = asyncio.create_task(self._key_send_task(key))

    def key_up(self, key=None):
        """
        Release a key that has been pressed down
        
        Args:
            key (str, optional): The key to release
                If no key is specified, all keys will be released
        """
        if key:
            self._key_cancel_task(key)
        else:
            for k in self.key_tasks.keys():
                self.key_up(k)

    def type_key(self, char):
        """
        Sends a key to the client.
        This is a different event than ``send_key`` and is only useful for the chat window.
        
        Args:
            char: The character to type
                "TAB", "ENTER", "ALT", "ESC", "SPACE" and others are also accepted as special keys
        """
        code = None
        try:
            code = keycode_map[char]
        except KeyError:
            code = ord(char)

        user32.PostMessageW(self.window_handle, 0x102, code, 0)

    def type_string(self, string):
        """
        Type a string of letters directly to the window.
        
        Args:
            string: the text to type
        """
        for s in string:
            self.type_key(s)

    async def _key_send_task(self, key):
        while True:
            self._send_key_event(key, 0)
            await asyncio.sleep(0.1)

    def _key_cancel_task(self, key):
        if key in self.key_tasks.keys():
            self.key_tasks[key].cancel()

        self._send_key_event(key, 1)

    def _send_key_event(self, key, event):
        try:
            code = keycode_map[key.upper()]
            msg = 0x101 if event else 0x100
            # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-sendmessagew
            # https://docs.microsoft.com/en-us/windows/win32/inputdev/wm-keydown
            user32.PostMessageW(self.window_handle, msg, code, 0)
        except KeyError:
            print("Invalid key provided")


if __name__ == "__main__":
    """ Some tests """
    from wizwalker.utils import get_all_wizard_handles

    try:
        window_handle = get_all_wizard_handles()[0]
    except IndexError:
        print("No running wizard101 windows")
        exit(0)

    keyboard = Keyboard(window_handle)

    async def say_hello():
        keyboard.type_key("\r")
        await asyncio.sleep(0.2)
        keyboard.type_string("Hello world!\r")

    async def test():
        # Hold down 2 keys
        keyboard.key_down("D")
        keyboard.key_down("W")
        await asyncio.sleep(2)
        # Release 1
        keyboard.key_up("D")
        await asyncio.sleep(1)
        # Release all keys
        keyboard.key_up()
        # Hold s for .5 seconds
        await keyboard.hold_key("s", 0.5)

        # Wait and say hello!
        await asyncio.sleep(0.1)
        await say_hello()

    asyncio.run(test())



# Native imports
import ctypes
from ctypes.wintypes import POINT
import time

# Custom imports
from wizsdk.window import Window, screen_size

# If the mouse is over a coordinate in FAILSAFE_POINTS and FAILSAFE is True, the FailSafeException is raised.
# The rest of the points are added to the FAILSAFE_POINTS list at the bottom of this file, after size() has been defined.
# The points are for the corners of the screen, but note that these points don't automatically change if the screen resolution changes.
FAILSAFE = True
FAILSAFE_POINTS = [(0, 0)]
MINIMUM_DURATION = 0.1
MINIMUM_SLEEP = 0.02


class FailSafeException(Exception):
    """
    Exception raised when the mouse is over (0, 0). Protactive measure to regain control of your mouse.
    """

    def __init__(
        self,
        message="Mouse fail-safe triggered from mouse moving to a corner of the screen.",
    ):
        self.message = message
        super().__init__(self.message)


def getPointOnLine(x1, y1, x2, y2, n):
    """
    Returns an (x, y) tuple of the point that has progressed a proportion ``n`` along the line defined by the two
    ``x1``, ``y1`` and ``x2``, ``y2`` coordinates.
    This function was copied from pytweening module, so that it can be called even if PyTweening is not installed.
    """
    x = ((x2 - x1) * n) + x1
    y = ((y2 - y1) * n) + y1
    return (x, y)


class Mouse(Window):
    """
    Class for controlling the computer's mouse.
    """

    _MOUSEEVENTF_MOVE = 0x0001  # mouse move
    _MOUSEEVENTF_LEFTDOWN = 0x0002  # left button down
    _MOUSEEVENTF_LEFTUP = 0x0004  # left button up
    _MOUSEEVENTF_RIGHTDOWN = 0x0008  # right button down
    _MOUSEEVENTF_RIGHTUP = 0x0010  # right button up
    _MOUSEEVENTF_MIDDLEDOWN = 0x0020  # middle button down
    _MOUSEEVENTF_MIDDLEUP = 0x0040  # middle button up
    _MOUSEEVENTF_WHEEL = 0x0800  # wheel button rolled
    _MOUSEEVENTF_ABSOLUTE = 0x8000  # absolute move
    _SM_CXSCREEN = 0
    _SM_CYSCREEN = 1

    def __init__(self, window_handle=None):
        super().__init__(window_handle)
        # Window handle to which the mouse events will be relative to
        self.window_handle = window_handle

    def _do_event(self, flags, x_pos, y_pos, data, extra_info):
        """generate a mouse event"""
        x_calc = int(
            65536 * x_pos / ctypes.windll.user32.GetSystemMetrics(self._SM_CXSCREEN) + 1
        )
        y_calc = int(
            65536 * y_pos / ctypes.windll.user32.GetSystemMetrics(self._SM_CYSCREEN) + 1
        )
        return ctypes.windll.user32.mouse_event(flags, x_calc, y_calc, data, extra_info)

    def _get_button_value(self, button_name, button_up=False):
        """convert the name of the button into the corresponding value"""
        buttons = 0
        if button_name.find("right") >= 0:
            buttons = self._MOUSEEVENTF_RIGHTDOWN
        if button_name.find("left") >= 0:
            buttons = buttons + self._MOUSEEVENTF_LEFTDOWN
        if button_name.find("middle") >= 0:
            buttons = buttons + self._MOUSEEVENTF_MIDDLEDOWN
        if button_up:
            buttons = buttons << 1
        return buttons

    def _set_position(self, pos):
        """
        Set the position of the mouse to the specified coordinates
        relative to the window
        """
        (x, y) = pos

        old_pos = self.get_position()
        x = x if (x != -1) else old_pos[0]
        y = y if (y != -1) else old_pos[1]
        self._do_event(self._MOUSEEVENTF_MOVE + self._MOUSEEVENTF_ABSOLUTE, x, y, 0, 0)

    async def move_to(self, x, y, duration=0.5):
        """
        Move the mouse to the x, y coordinates relative to the window
        """
        # We need to get from (startx, starty) to (x, y)

        # Set the X, Y to be relative to the window position
        wX, wY, *_ = self.get_rect()
        x += wX
        y += wY

        startx, starty = self.get_position()
        x_offset = x - startx
        y_offset = y - starty

        width, height = screen_size()

        steps = [(x, y)]

        if duration > MINIMUM_DURATION:
            num_steps = max(width, height)
            sleep_amount = duration / num_steps

            if sleep_amount < MINIMUM_SLEEP:
                num_steps = int(duration / MINIMUM_SLEEP)
                sleep_amount = duration / num_steps

            steps = [
                getPointOnLine(startx, starty, x, y, (n / num_steps))
                for n in range(num_steps)
            ]
            steps.append((x, y))

        for _x, _y in steps:
            if len(steps) > 1:
                # A single step doesn't require tweening
                time.sleep(sleep_amount)

            _x = int(round(_x))
            _y = int(round(_y))

            # Failsafe check
            if (_x, _y) not in FAILSAFE_POINTS:
                self.failSafeCheck()

            self._set_position((_x, _y))

        # Failsafe check
        if (_x, _y) not in FAILSAFE_POINTS:
            self.failSafeCheck()

    def press_button(self, x=-1, y=-1, button="left", button_up=False):
        """push a button of the mouse"""
        self.move_to(x, y)
        self._do_event(self.get_button_value(button, button_up), 0, 0, 0, 0)

    async def click(self, x=-1, y=-1, button="left", duration=None, delay=0.1):
        """Click at the specified placed"""
        self.set_active()

        # Set default duration
        if duration is None:
            if x == -1 and y == -1:
                duration = 0
            else:
                duration = 0.5

        # If position is not set, use current mouse position
        old_pos = self.get_position()
        x = x if (x != -1) else old_pos[0]
        y = y if (y != -1) else old_pos[1]
        await self.move_to(x, y, duration=duration)
        time.sleep(delay)
        self._do_event(
            self._get_button_value(button, False)
            + self._get_button_value(button, True),
            0,
            0,
            0,
            0,
        )

    def double_click(self, pos=(-1, -1), button="left"):
        """Double click at the specifed placed"""
        for i in xrange(2):
            self.click(pos, button)

    def get_position(self):
        """get mouse position"""
        point = POINT()
        ctypes.windll.user32.GetCursorPos(ctypes.byref(point))
        return (point.x, point.y)

    def get_rel_position(self):
        """get mouse position relative to window"""
        wx, wy = self.get_rect()[:2]
        x, y = self.get_position()
        return (x - wx, y - wy)

    def in_rect(self, rect_area):
        """
        Returns true if the mouse is in the given region
        """
        x, y, w, h = rect_area
        wx, wy = self.get_rect()[:2]
        # Make the position of the rect relative to the window
        x += wx
        y += wy

        mouseX, mouseY = self.get_position()
        return mouseX > x and mouseX < (x + w) and mouseY > y and mouseY < (y + h)

    async def move_out(self, rect_area):
        """
        Move the mouse outside of the given rect
        If the mouse is already outside, return
        """
        x, y = self.get_rel_position()
        while self.in_rect(rect_area):
            # TODO find fastest way out of rect
            y -= 100
            await self.move_to(x, y, duration=0.2)

    def failSafeCheck(self):
        if FAILSAFE and self.get_position() in FAILSAFE_POINTS:
            raise FailSafeException


if __name__ == "__main__":
    import asyncio

    async def main():
        mouse = Mouse()
        mouse._set_position((100, 100))
        await mouse.move_to(10, 100, duration=1)

    asyncio.run(main())




# Native imports
import sys
from contextlib import suppress

# Third-party imports
import asyncio
from wizwalker import HookNotActive

# Custom imports
from .pixels import DeviceContext, match_image
from .card import Card
from .utils import packaged_img


class Battle(DeviceContext):
    """
    Battle class
    
    Example:
        .. code-block:: py
        
            # register the client
            p1 = Client.register(name="Bot")
            # Get a battle object from the client
            battle = p1.get_battle("Test")
            # Loop
            while await battle.loop():
                if battle.round_count == 1:
                    # pass on the first round
                    p1.pass_turn()
                else:
                    # Cast enchanted tempest
                    temp = await p1.find_spell("tempest")
                    epic = await p1.find_spell("epic")

                    if temp and epic:
                        e_temp = await epic.enchant(temp)
                        await e_temp.cast()
                    else:
                        await p1.pass_turn()
    """

    def __init__(self, client, name=None, default_image_folder=""):
        super().__init__(client.window_handle)
        self.client = client
        self._default_image_folder = default_image_folder

        self.logging = self.client.logging
        self.is_idle = self.client.is_idle
        self.name = name

        self._round_count = 1
        self._going_first = False
        self.enemy_first = False

        self.is_over = False

        self.in_progress = False

        # rectangles defined as (x, y, width, height)
        self._spell_area = (245, 290, 370, 70)
        self._enemy_area = (68, 26, 650, 50)
        self._ally_area = (140, 580, 650, 50)

    @property
    def round_count(self) -> int:
        """
        Current round of the fight. Starts at 1
        """
        return self._round_count

    @property
    def going_first(self) -> bool:
        """
        True if your team is going first, False if enemies are going first. Value is set once at the beginning of the fight.
        """
        return self._going_first

    async def loop(self):
        """
        Handles the looping logic for a battle. The loops exists when the battle ends.
        
        Examples:
            .. code-block:: py
                
                while await battle.loop():
                    print(battle.round_count)
                    print(battle.enemy_first)
                    print(battle.get_enemy_count())
        """

        if not self.in_progress and not self.is_over:
            await self._start()
            return True
        else:
            await self._next_turn()
            return not self.is_over

    def log(self, message):
        if self.logging:
            s = ""
            if self.name != None:
                s += f"[{self.name}] "
            s += message
            print(s)
            sys.stdout.flush()

    def print_round(self):
        self.log(f"----------- Battle round {self._round_count} -----------")

    def _is_turn(self) -> bool:
        """
        Returns if it's our turn to play
        by matching pixels in the `flee` button
        """
        flee_yellow = self.pixel_matches_color((550, 400), (255, 255, 0), tolerance=30)
        flee_brown = self.pixel_matches_color((578, 394), (124, 68, 0), tolerance=30)
        return flee_yellow and flee_brown

    async def _start(self) -> None:
        """
        Waits for the first round then signals to the class that the battle has started.
        used in the `loop()` method
        """
        while not self._is_turn():
            await asyncio.sleep(1)

        self.is_over = False
        self.in_progress = True

        self.enemy_first = self._is_enemy_first()
        self._going_first = not self.enemy_first

        who_is = "Enemy is" if self.enemy_first else "You are"

        self.log(">> Battle is starting")
        self.log("======================================")
        self.log(f"{who_is} going first")

        self.print_round()

    async def _next_turn(self) -> None:
        """
        Sleeps until next round or fight has ended
        Sets `is_over` to True if the fight has ended
        Increase the `_round_count` otherwise
        Used in the `loop()` method
        """
        while self._is_turn():
            await asyncio.sleep(1)

        while not self._is_turn() and not self.is_idle():
            await asyncio.sleep(1)

        if self.is_idle():
            self.log("Battle has finished")

            with suppress(HookNotActive):
                await self.client.get_mana()
                await self.client.get_health()

            await asyncio.sleep(0.5)
            self.is_over = True
            self.in_progress = False
        else:
            self._round_count += 1
            self.print_round()

    def get_enemy_positions(self):
        """
        Gets the indices of the enemies present
        
        Examples:
            .. code-block:: py
            
                # Register
                player = client.register(name="test")
                # Get the battle
                battle = player.get_battle()
                while await battle.loop():
                    enemies = battle.get_enemy_positions() # -> [0, 1]
                    print(len(enemies)) # -> 2
                    # Get the position of the first enemy in battle
                    first_enemy = enemies[0] # -> 0
                    # Get spell
                    firecat = await player.find_spell('firecat') # -> Card
                    if firecat:
                        # Cast at enemy position
                        await firecat.cast(target=first_enemy)
        """
        Y = 75
        COLOR = (207, 186, 135)
        enemies = []
        for i in range(4):
            X = (174 * (i)) + 203
            if self.pixel_matches_color((X, Y), COLOR, tolerance=30):
                enemies.append(i)

        return enemies

    def get_enemy_count(self):
        """
        Returns the number enemies in the fight
        """
        return len(self.get_enemy_positions())

    def find_enemy(self, enemy_image):
        """ 
        Attemps to find the position of an enemy that matches the image provided 
        returns 0, 1, 2, 3 if found otherwise returns False
        """
        found = self.locate_on_screen(
            enemy_image, region=self._enemy_area, threshold=0.2
        )

        if found:
            return round((found[0] - 60) / 170)

        return False

    def find_ally(self, ally_image):
        """ 
        Attemps to find the position of an ally the matches the image provided 
        returns 4, 5, 6, 7 if found otherwise returns False
        """

        found = self.locate_on_screen(ally_image, region=self._ally_area, threshold=0.2)

        if found:
            return 7 - round((found[0] - 100) / 170)

        return False

    def _is_enemy_first(self):
        turn_arrow_region = (230, 240, 80, 60)
        return bool(
            self.client.locate_on_screen(
                "enemy-first.png",
                region=turn_arrow_region,
                threshold=0.2,
                folder=packaged_img(),
            )
        )

